/* a2dos33_img.h version 1.0.0 */
#ifndef __A2DOS33_IMG_H__
#define __A2DOS33_IMG_H__

#include "a2dsk_img.h"

#define A2DOS33_SECTOR_SIZE				(256)

#define A2DOS33_VTOC_TRACK				(0x11)
#define A2DOS33_VTOC_SECTOR				(0x00)

#define A2DOS33_CATALOG_START_TRACK		(0x11)
#define A2DOS33_CATALOG_START_SECTOR	(0x0F)

#define A2DOS33_NUM_TS_LIST				(122)
#define A2DOS33_NUM_TRACKS				(50)
#define A2DOS33_NUM_CATALOG_ENTRY		(7)
#define A2DOS33_FILENAME_LEN			(30)

struct a2dos33_track_sector_t {
	unsigned char track;
	unsigned char sector;
};

struct a2dos33_ts_list_t {
	struct a2dos33_ts_list_t *next;

	unsigned char unused1;
	struct a2dos33_track_sector_t next_ts_list_sector;
	unsigned char unused2[2];
	unsigned short sector_number_in_file;
	unsigned char unused3[5];
	struct a2dos33_track_sector_t list[A2DOS33_NUM_TS_LIST];
};

struct a2dos33_catalog_entry_t {
	struct a2dos33_ts_list_t *ts_list;

	struct a2dos33_track_sector_t first_ts_list_sector;
	unsigned char file_type;
	char filename[A2DOS33_FILENAME_LEN];
	unsigned short num_sectors;
};

struct a2dos33_catalog_t {
	struct a2dos33_catalog_t *next;

	unsigned char unused1;
	struct a2dos33_track_sector_t next_catalog_sector;
	unsigned char unused2[8];
	struct a2dos33_catalog_entry_t entry[A2DOS33_NUM_CATALOG_ENTRY];
};

struct a2dos33_vtoc_t {
	unsigned char unused1;
	struct a2dos33_track_sector_t catalog_sector;
	unsigned char dos_version;
	unsigned char unused2[2];
	unsigned char disk_volume;
	unsigned char unused3[32];
	unsigned char num_ts_pairs;
	unsigned char unused4[8];
	unsigned char last_track;
	unsigned char dir_alloc;
	unsigned char unused5[2];
	unsigned char num_tracks_per_disk;
	unsigned char num_sectors_per_track;
	unsigned short num_bytes_per_sector;
	unsigned int track_bitmap[A2DOS33_NUM_TRACKS];
};

struct a2dos33_image_t {
	struct a2dos33_vtoc_t vtoc;
	struct a2dos33_catalog_t *catalog;
	const int *sector_order_table;

	int sector_order;
	int disk_size;
	int num_tracks;
	unsigned char *disk_img;
};

extern void free_a2dos33_image(struct a2dos33_image_t *image, int free_all);
extern int read_a2dos33_image(struct a2dos33_image_t *image, const char *image_name, int sector_order);
extern int read_a2dos33_image_file(const struct a2dos33_image_t *image, const char *filename, unsigned char *data, int data_size);

#endif	/* __A2DOS33_IMG_H__ */
