#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bitmap.h"
#include "a2dos33_img.h"


static const struct bitmap_palette_t apple2_palette[16] = {
	RGB(0x00,0x00,0x00),		/* black */
#if 1
	RGB(0xD0,0x50,0xD0),		/* pink */
	RGB(0x60,0xD0,0x50),		/* green */
	RGB(0x50,0xA0,0xFF),		/* blue */
	RGB(0xE0,0x80,0x20),		/* orange */
#else
#if 1
	RGB(0xC0,0x40,0xD0),		/* pink */
	RGB(0x40,0xC0,0x30),		/* green */
	RGB(0x20,0x90,0xF0),		/* blue */
	RGB(0xE0,0x70,0x10),		/* orange */
#else
	RGB(0xD0,0x50,0xD0),		/* pink */
	RGB(0x60,0xD0,0x50),		/* green */
	RGB(0x50,0xA0,0xFF),		/* blue */
	RGB(0xE0,0x80,0x20),		/* orange */
#endif
#endif
	RGB(0xFF,0xFF,0xFF),		/* white */
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00)
};

static const struct bitmap_palette_t digital_palette[16] = {
	RGB(0x00,0x00,0x00),		/* black   */
	RGB(0x00,0x00,0xFF),		/* blue    */
	RGB(0xFF,0x00,0x00),		/* red     */
	RGB(0xFF,0x00,0xFF),		/* magenta */
	RGB(0x00,0xFF,0x00),		/* green   */
	RGB(0x00,0xFF,0xFF),		/* cyan    */
	RGB(0xFF,0xFF,0x00),		/* yellow  */
	RGB(0xFF,0xFF,0xFF),		/* white   */
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00)
};

static const unsigned char dither_table[6][2] = {
	{ 0, 0 },	/* black           */
#if 1
	{ 3, 1 },	/* magenta & blue  */
#else
	{ 2, 1 },	/* red     & blue  */
#endif
#if 1
	{ 4, 4 },	/* green           */
#else
	{ 5, 6 },	/* cyan  & yellow  */
#endif
	{ 5, 1 },	/* cyam  & blue    */
	{ 2, 6 },	/* red   & yellow  */
	{ 7, 7 }	/* white           */
};

static const struct bitmap_palette_t ega_palette[16] = {
	RGB(0x00, 0x00, 0x00),		/*  0 black         */
	RGB(0x00, 0x00, 0xAA),		/*  1 blue          */
	RGB(0x00, 0xAA, 0x00),		/*  2 green         */
	RGB(0x00, 0xAA, 0xAA),		/*  3 cyan          */
	RGB(0xAA, 0x00, 0x00),		/*  4 red           */
	RGB(0xAA, 0x00, 0xAA),		/*  5 magenta       */
	RGB(0xAA, 0x55, 0x00),		/*  6 brown         */
	RGB(0xAA, 0xAA, 0xAA),		/*  7 gray          */
	RGB(0x55, 0x55, 0x55),		/*  8 dark gray     */
	RGB(0x55, 0x55, 0xFF),		/*  9 light blue    */
	RGB(0x55, 0xFF, 0x55),		/* 10 light green   */
	RGB(0x55, 0xFF, 0xFF),		/* 11 light cyan    */
	RGB(0xFF, 0x55, 0x55),		/* 12 light red     */
	RGB(0xFF, 0x55, 0xFF),		/* 13 light magenta */
	RGB(0xFF, 0xFF, 0x55),		/* 14 light yellow  */
	RGB(0xFF, 0xFF, 0xFF)		/* 15 white         */
};

static const unsigned char ega_color[6] = {
	 0,			/* black       */
	 5,			/* magenta     */
#if 1
	 2,			/* green       */
#else
	10,			/* light green */
#endif
	 9,			/* light blue  */
	12,			/* light red   */
	15,			/* white       */
};

static const unsigned char dither_table_ega[6][2] = {
	{  0,  0 },		/* black        */
	{  5, 13 },		/* magenta      */
	{ 10,  2 },		/* green        */
#if 1
	{  3,  9 },		/* cyan & blue  */
#else
	{  9,  1 },		/* blue         */
#endif
#if 1
	{  6, 12 },		/* red & brown  */
#else
	{ 14,  4 },		/* red & yellow */
#endif
	{ 15, 15 }		/* white        */
};


#define COLOR_MODE_MONO			(0)
#define COLOR_MODE_NO_WHITE		(1)
#define COLOR_MODE_BLUR			(2)
#define COLOR_MODE_NO_BLUR		(3)


#if 0
static const int image_offset[2][17] = {
	{
		/* SHP0 */
		0x13100, 0x13000, 0x14F00, 0x14E00, 0x14D00, 0x14C00, 0x14B00, 0x14A00,
		0x14900, 0x14800, 0x14700, 0x14600, 0x14500, 0x14400, 0x14300, 0x14200,
		0x14100
	},
	{
		/* SHP1 */
		0x15F00, 0x15E00, 0x15D00, 0x15C00, 0x15B00, 0x15A00, 0x15900, 0x15800,
		0x15700, 0x15600, 0x15500, 0x15400, 0x15300, 0x15200, 0x15100, 0x15000,
		0x16F00
	}
};
#endif

static void usage(void);
static void convert_to_ega_color(unsigned char *vram, int vram_size);
static void convert_to_dither_color(unsigned char *vram, int width, int height, const unsigned char table[6][2]);
static void convert_to_dither_color2(unsigned char *vram, int width, int height, const unsigned char table[6][2]);


static void usage(void)
{
	printf("usage: u4a2shapes [options] OUTPUT.BMP\n");
	printf("option: --disk DISK-IMG     read shape data from specified image\n");
	printf("                            read 'U4boot.dsk' if not specified\n");
	printf("        --do | --po         specify sector order of disk image\n");
	printf("                                --do    DOS 3.3 order\n");
	printf("                                --po    ProDOS order\n");
	printf("        --16                draw shape in 16 pixel width\n");
	printf("        --2                 scale x2\n");
	printf("        --half-shift        half shift blur and orange pixel when scale x2 mode\n");
	printf("        --dither            dither graphics (stripe)         when scale x2 mode\n");
	printf("        --dither2           dither graphics (cross)          when scale x2 mode\n");
	printf("        --ega               use ega color\n");
	printf("        --mono              monochrome graphics\n");
	printf("        --no-white          do not blend green/pink pair and orange/blue pair\n");
	printf("                            to white\n");
	printf("        --blur              blur graphics\n");
	printf("        --no-blur           do not blur\n");
	printf("        --help              print usage\n");
}

int main(int argc, char **argv)
{
	int num_cmd_params = 1;
	const char *bmp_file = NULL;
	const char *image_file = "U4boot.dsk";
	const char *shp_file[2] = { "S\x01HP0", "S\x01HP1" };
	int shape_width = 14;
	int shape_height = 16;
	int scale = 1;
	int half_shift = 0;
	int opt_dither = 0;
	int opt_ega = 0;
	int color_mode = COLOR_MODE_BLUR;
	unsigned char shp_data[2][0x1000];
	unsigned char blend_buf[32];
	unsigned char *vram = NULL;
	int vram_size;
	int row;
	int col;
	int scale_height;
	int scale_width;
	int half_scale_width;
	int h_margin;
	struct a2dos33_image_t image;
	struct bitmap_header_t header;
	int sector_order = A2DSK_SECTOR_ORDER_UNKNOWN;
	FILE *fp;
	int i;

	while (*++argv) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--disk") == 0) {
				if (*++argv == NULL) {
					printf("no disk image filename\n");
					exit(1);
				}
				image_file = *argv;
			} else if (strcmp(*argv, "--do") == 0) {
				sector_order = A2DSK_SECTOR_ORDER_DOS33;
			} else if (strcmp(*argv, "--po") == 0) {
				sector_order = A2DSK_SECTOR_ORDER_PRODOS;
			} else if (strcmp(*argv, "--16") == 0) {
				shape_width = 16;
			} else if (strcmp(*argv, "--2") == 0) {
				scale = 2;
			} else if (strcmp(*argv, "--half-shift") == 0) {
				half_shift = 1;
			} else if (strcmp(*argv, "--dither") == 0) {
				opt_dither = 1;
			} else if (strcmp(*argv, "--dither2") == 0) {
				opt_dither = 2;
			} else if (strcmp(*argv, "--ega") == 0) {
				opt_ega = 1;
			} else if (strcmp(*argv, "--mono") == 0) {
				color_mode = COLOR_MODE_MONO;
			} else if (strcmp(*argv, "--no-white") == 0) {
				color_mode = COLOR_MODE_NO_WHITE;
			} else if (strcmp(*argv, "--blur") == 0) {
				color_mode = COLOR_MODE_BLUR;
			} else if (strcmp(*argv, "--no-blur") == 0) {
				color_mode = COLOR_MODE_NO_BLUR;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			bmp_file = *argv;
			num_cmd_params--;
		}
	}

	if (scale != 2) {
		/* reset options */
		half_shift = 0;
		opt_dither = 0;
	}

	if (bmp_file == NULL) {
		printf("no output bitmap filename\n");
		return 1;
	}

	scale_height = shape_height * scale;
	scale_width = shape_width * scale;
	half_scale_width = scale_width / 2;
	h_margin = (shape_width - 14) * scale / 2;

	if (read_a2dos33_image(&image, image_file, sector_order) < 0) {
		printf("read disk image error '%s'\n", image_file);
		return 1;
	}

	for (i = 0; i < 2; i++) {
		int read_size;
		if ((read_size = read_a2dos33_image_file(&image, shp_file[i], shp_data[i], sizeof(shp_data[0]))) < 0) {
			printf("read disk image error '%s/%s'\n", image_file, shp_file[i]);
			return 1;
		}
		if (read_size != sizeof(shp_data[0])) {
			printf("read disk image error '%s/%s' size=0x%X\n", image_file, shp_file[i], read_size);
			return 1;
		}
	}

	free_a2dos33_image(&image, 0);

	vram_size = scale_width * scale_height * 16 * 16;
	vram = (unsigned char *)malloc(vram_size);
	if (vram == NULL) {
		printf("memory exhaust!\n");
		exit(1);
	}

	memset(vram, 0, vram_size);

	for (row = 0; row < 16; row++) {
		for (col = 0; col < 16; col++) {
			int y;

			for (y = 0; y < shape_height; y++) {
				const char mask[7] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40 };
				const char color[3][8] = {
					{ 2, 1, 2, 1, 2, 1, 2, 1 },
					{ 4, 3, 4, 3, 4, 3, 4, 3 },
					{ 5, 5, 5, 5, 5, 5, 5, 5 }
				};
				int x = h_margin;
				int i;
				unsigned char data;

				memset(blend_buf, 0, sizeof(blend_buf));

				data = shp_data[0][row * 16 + col + y * 256];
				if (data & 0x80) {
					const char *c = (color_mode == COLOR_MODE_MONO ? color[2] : color[1]);

					if (half_shift) {
						x++;
					}

					for (i = 0; i < 7; i++, x += scale) {
						if (data & mask[i]) {
							blend_buf[x] = c[i];
							if (scale == 2 && x < half_scale_width) {
								blend_buf[x + 1] = c[i];
							}
						}
					}
				} else {
					const char *c = (color_mode == COLOR_MODE_MONO ? color[2] : color[0]);

					for (i = 0; i < 7; i++, x += scale) {
						if (data & mask[i]) {
							blend_buf[x] = c[i];
							if (scale == 2) {
								blend_buf[x + 1] = c[i];
							}
						}
					}
				}

				data = shp_data[1][row * 16 + col + y * 256];
				x = half_scale_width;
				if (data & 0x80) {
					const char *c = (color_mode == COLOR_MODE_MONO ? color[2] : (color[1] + 1));

					if (half_shift) {
						if (blend_buf[x - 1]) {
							blend_buf[x] = c[1];
						}
						x++;
					}

					for (i = 0; i < 7; i++, x += scale) {
						if (data & mask[i]) {
							blend_buf[x] = c[i];
							if (scale == 2 && x < scale_width) {
								blend_buf[x + 1] = c[i];
							}
						}
					}
				} else {
					const char *c = (color_mode == COLOR_MODE_MONO ? color[2] : (color[0] + 1));

					for (i = 0; i < 7; i++, x += scale) {
						if (data & mask[i]) {
							blend_buf[x] = c[i];
							if (scale == 2) {
								blend_buf[x + 1] = c[i];
							}
						}
					}
				}

				if (color_mode == COLOR_MODE_BLUR || color_mode == COLOR_MODE_NO_BLUR) {
					if (scale == 1) {
						if (blend_buf[0] && blend_buf[1]) {
							blend_buf[0] = 5;
						}
						for (x = 1; x < scale_width - 1; x++) {
							if (blend_buf[x]) {
								if (blend_buf[x - 1] || blend_buf[x + 1]) {
									blend_buf[x] = 5;
								}
							}
						}
						if (blend_buf[scale_width - 2] && blend_buf[scale_width - 1]) {
							blend_buf[scale_width - 1] = 5;
						}

						if (color_mode == COLOR_MODE_BLUR) {
							for (x = 1; x < scale_width - 1; x++) {
								if (! blend_buf[x]) {
									if (blend_buf[x - 1] && blend_buf[x + 1]) {
										if (blend_buf[x - 1] != 5) {
											if (blend_buf[x + 1] != 5) {
												blend_buf[x] = (x < half_scale_width) ? blend_buf[x - 1] : blend_buf[x + 1];
											} else {
												blend_buf[x] = blend_buf[x - 1];
											}
										} else {
											if (blend_buf[x + 1] != 5) {
												blend_buf[x] = blend_buf[x + 1];
											}
										}
									}
								}
							}
						}
					} else if (scale == 2) {
						for (x = 1; x < scale_width - 1; x++) {
							if (blend_buf[x - 1] && blend_buf[x] && blend_buf[x + 1]) {
								blend_buf[x] = 5;
							}
						}

						if (color_mode == COLOR_MODE_BLUR) {
							unsigned char blend_buf2[32];

							memcpy(blend_buf2, blend_buf, sizeof(blend_buf));

							if (! blend_buf2[0] && blend_buf2[1]) {
								blend_buf[0] = blend_buf2[1];
							}

							for (x = 1; x < scale_width - 1; x++) {
								if (! blend_buf2[x]) {
									if (blend_buf2[x - 1]) {
										if (blend_buf2[x + 1]) {
											blend_buf[x] = (x < half_scale_width) ? blend_buf2[x - 1] : blend_buf2[x + 1];
										} else {
											blend_buf[x] = blend_buf2[x - 1];
										}
									} else {
										if (blend_buf2[x + 1]) {
											blend_buf[x] = blend_buf2[x + 1];
										}
									}
								}
							}

							if (! blend_buf2[scale_width - 1] && blend_buf2[scale_width - 2]) {
								blend_buf[scale_width - 1] = blend_buf2[scale_width - 2];
							}

							if (blend_buf2[0] && blend_buf2[0] != 5 && blend_buf2[1] == 5) {
								blend_buf[0] = 5;
							}

							for (x = 1; x < scale_width - 1; x++) {
								if (blend_buf2[x] && blend_buf2[x] != 5) {
									if (blend_buf2[x - 1] == 5 || blend_buf2[x + 1] == 5) {
										blend_buf[x] = 5;
									}
								}
							}

							if (blend_buf2[scale_width - 1] && blend_buf2[scale_width - 1] != 5 && blend_buf2[scale_width - 2] == 5) {
								blend_buf[scale_width - 1] = 5;
							}
						}
					}
				}

				for (i = 0; i < scale; i++) {
					unsigned char *des = vram + (row * scale_height + y * scale + i) * (16 * scale_width) + (col * scale_width);
					const unsigned char *src = blend_buf;
					int length = scale_width;
					while (length--) {
						*des++ = *src++;
					}
				}
			}
		}
	}

	if (color_mode != COLOR_MODE_MONO) {
		if (opt_ega) {
			if (opt_dither == 1) {
				convert_to_dither_color(vram, scale_width * 16, scale_height * 16, dither_table_ega);
			} else if (opt_dither == 2) {
				convert_to_dither_color2(vram, scale_width * 16, scale_height * 16, dither_table_ega);
			} else {
				convert_to_ega_color(vram, vram_size);
			}
		} else {
			if (opt_dither == 1) {
				convert_to_dither_color(vram, scale_width * 16, scale_height * 16, dither_table);
			} else if (opt_dither == 2) {
				convert_to_dither_color2(vram, scale_width * 16, scale_height * 16, dither_table);
			}
		}
	}

	fp = fopen(bmp_file, "wb");
	if (fp == NULL) {
		printf("open error '%s'\n", bmp_file);
		return 1;
	}
	create_bitmap_header(&header, scale_width * 16, scale_height * 16, 4);
	write_bitmap(fp, &header, (opt_ega ? ega_palette : (opt_dither ? digital_palette : apple2_palette)), vram);
	fclose(fp);

	free(vram);

	return 0;
}

static void convert_to_ega_color(unsigned char *vram, int vram_size)
{
	while (0 < vram_size--) {
		*vram = ega_color[*vram];
		vram++;
	}
}

static void convert_to_dither_color(unsigned char *vram, int width, int height, const unsigned char table[6][2])
{
	int x;
	int y;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			*vram = table[*vram][x & 1];
			vram++;
		}
	}
}

static void convert_to_dither_color2(unsigned char *vram, int width, int height, const unsigned char table[6][2])
{
	int x;
	int y;

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			*vram = table[*vram][(x + y) & 1];
			vram++;
		}
	}
}

/* EOF */
