#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bitmap.h"


static const struct bitmap_palette_t palette[16] = {
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0xAA),
	RGB(0x00, 0xAA, 0x00),
	RGB(0x00, 0xAA, 0xAA),
	RGB(0xAA, 0x00, 0x00),
	RGB(0xAA, 0x00, 0xAA),
	RGB(0xAA, 0x55, 0x00),
	RGB(0xAA, 0xAA, 0xAA),
	RGB(0x55, 0x55, 0x55),
	RGB(0x55, 0x55, 0xFF),
	RGB(0x55, 0xFF, 0x55),
	RGB(0x55, 0xFF, 0xFF),
	RGB(0xFF, 0x55, 0x55),
	RGB(0xFF, 0x55, 0xFF),
	RGB(0xFF, 0xFF, 0x55),
	RGB(0xFF, 0xFF, 0xFF)
};

static void usage(void);



static void usage(void)
{
	printf("usage: u4ega_shapes [options] SHAPES.EGA OUTPUT.BMP\n");
	printf("option: --help             print usage\n");
}

int main(int argc, char **argv)
{
	int num_cmd_params = 2;

	const char *shp_file = NULL;		/* SHAPES.EGA */
	const char *bmp_file = NULL;
	unsigned char data[32768];
	FILE *fp;
	unsigned char vram[256 * 256];
	int row;
	int col;
	const unsigned char *src;
	struct bitmap_header_t header;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				shp_file = *argv;
			} else {
				bmp_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (shp_file == NULL) {
		printf("no shapes filename\n");
		return 1;
	}
	if (bmp_file == NULL) {
		printf("no output bitmap filename\n");
		return 1;
	}

	fp = fopen(shp_file, "rb");
	if (fp == NULL) {
		printf("open error '%s'\n", shp_file);
		return 1;
	}

	if (fread(data, sizeof(data), 1, fp) != 1) {
		fclose(fp);
		printf("read error '%s'\n", shp_file);
		return 1;
	}

	fclose(fp);

	memset(vram, 0, sizeof(vram));

	src = data;
	for (row = 0; row < 16; row++) {
		for (col = 0; col < 16; col++) {
			int y;
			int x;
			for (y = 0; y < 16; y++) {
				unsigned char *des = vram + (row * 16 + y) * 256 + col * 16;
				for (x = 0; x < 16; x += 2) {
					unsigned char color0 = *src++;
					unsigned char color1 = color0 & 0x0F;
					color0 >>= 4;
					*des++ = color0;
					*des++ = color1;
				}
			}
		}
	}

	fp = fopen(bmp_file, "wb");
	if (fp == NULL) {
		printf("open error '%s'\n", bmp_file);
		return 1;
	}
	create_bitmap_header(&header, 256, 256, 4);
	write_bitmap(fp, &header, palette, vram);
	fclose(fp);

	return 0;
}

/* EOF */
